#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2025, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the SCC writer"""

# pylint: disable=R0201,C0115,C0116,W0212

import json
import os
import unittest
import xml.etree.ElementTree as et
from fractions import Fraction
from pathlib import Path

import sys
from ttconv.imsc.attributes import TimeExpressionSyntaxEnum
from ttconv.imsc.config import IMSCWriterConfiguration
import ttconv.imsc.reader as imsc_reader
import ttconv.imsc.writer as imsc_writer
import ttconv.scc.writer as scc_writer
import ttconv.scc.reader as scc_reader
from ttconv.scc.config import SCCFrameRate, SccWriterConfiguration
from ttconv.model import ContentDocument, Region, Body, Div, P, Span, Text, ContentElement
from ttconv.style_properties import StyleProperties, DisplayType

class SccWriterConfigurationTest(unittest.TestCase):

  def test_defaults(self):
    config = SccWriterConfiguration()

    self.assertEqual(config.allow_reflow, True)
    self.assertEqual(config.force_popon, False)
    self.assertEqual(config.frame_rate, SCCFrameRate.FPS_2997_DF)

  def test_allow_reflow(self):
    config = SccWriterConfiguration.parse(json.loads("""{"allow_reflow": true }"""))
    self.assertEqual(config.allow_reflow, True)

    config = SccWriterConfiguration.parse(json.loads("""{"allow_reflow": false }"""))
    self.assertEqual(config.allow_reflow, False)

  def test_force_popon(self):
    config = SccWriterConfiguration.parse(json.loads("""{"force_popon": true }"""))
    self.assertEqual(config.force_popon, True)

    config = SccWriterConfiguration.parse(json.loads("""{"force_popon": false }"""))
    self.assertEqual(config.force_popon, False)

  def test_rollup_lines(self):
    config = SccWriterConfiguration.parse(json.loads("""{"rollup_lines": 2 }"""))
    self.assertEqual(config.rollup_lines, 2)

    config = SccWriterConfiguration.parse(json.loads("""{"rollup_lines": 3 }"""))
    self.assertEqual(config.rollup_lines, 3)

    config = SccWriterConfiguration.parse(json.loads("""{"rollup_lines": 4 }"""))
    self.assertEqual(config.rollup_lines, 4)

    with self.assertRaises(ValueError):
      config = SccWriterConfiguration.parse(json.loads("""{"rollup_lines": 5 }"""))

  def test_frame_rate(self):
    config = SccWriterConfiguration.parse(json.loads("""{"frame_rate": "30NDF" }"""))
    self.assertEqual(config.frame_rate.fps, Fraction(30))
    self.assertFalse(config.frame_rate.df)

    config = SccWriterConfiguration.parse(json.loads("""{"frame_rate": "29.97NDF" }"""))
    self.assertEqual(config.frame_rate.fps, Fraction(30000, 1001))
    self.assertFalse(config.frame_rate.df)

    config = SccWriterConfiguration.parse(json.loads("""{"frame_rate": "29.97DF" }"""))
    self.assertEqual(config.frame_rate.fps, Fraction(30000, 1001))
    self.assertTrue(config.frame_rate.df)

    with self.assertRaises(ValueError):
      config = SccWriterConfiguration.parse(json.loads("""{"frame_rate": 30 }"""))

class SCCWriterTest(unittest.TestCase):

  def test_basic(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:frameRate="30" ttp:frameRateMultiplier="1000 1001">
  <body>
    <div>
      <p begin="30f" end="90f">Hello</p>
      <p begin="120f" end="150f">Bonjour</p>
     </div>
  </body>
</tt>"""

    expected_scc="""Scenarist_SCC V1.0

00:00:00;21	9420 9420 94ae 94ae 9440 9440 c8e5 ecec ef80 942f 942f

00:00:03;00	942c 942c

00:00:03;20	9420 9420 94ae 94ae 9440 9440 c2ef 6eea ef75 f280 942f 942f

00:00:05;00	942c 942c"""

  def test_pop_on_centered(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:frameRate="30" ttp:frameRateMultiplier="1000 1001">
  <body>
    <div>
      <p begin="30f" end="90f" tts:textAlign="center">Hello</p>
     </div>
  </body>
</tt>"""

    expected_scc="""Scenarist_SCC V1.0

00:00:00;21	9420 9420 94ae 94ae 94d6 94d6 20c8 e5ec ecef 942f 942f

00:00:03;00	942c 942c"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc_str)))
    assert model is not None
    config = SccWriterConfiguration()
    scc_from_model = scc_writer.from_model(model, config)
    self.assertEqual(scc_from_model, expected_scc)

  def test_pop_on_right_aligned(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:frameRate="30" ttp:frameRateMultiplier="1000 1001">
  <body>
    <div>
      <p begin="30f" end="90f" tts:textAlign="right">Hello</p>
     </div>
  </body>
</tt>"""

    expected_scc="""Scenarist_SCC V1.0

00:00:00;20	9420 9420 94ae 94ae 94dc 94dc 2020 20c8 e5ec ecef 942f 942f

00:00:03;00	942c 942c"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc_str)))
    assert model is not None
    config = SccWriterConfiguration()
    scc_from_model = scc_writer.from_model(model, config)
    self.assertEqual(scc_from_model, expected_scc)

  def test_rollup(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:frameRate="30" ttp:frameRateMultiplier="1000 1001">
  <body>
    <div>
      <p begin="30f" end="90f">Hello</p>
      <p begin="90f" end="150f">Hello my name</p>
      <p begin="150f" end="300f">Hello my name<br/>is Paul.</p>
     </div>
  </body>
</tt>"""

    expected_scc="""Scenarist_SCC V1.0

00:00:00;28	94a7 94a7 94ad 94ad 9470 9470 c8e5 ecec ef80

00:00:03;00	206d 7920 6e61 6de5

00:00:04;28	94a7 94a7 94ad 94ad 9470 9470 e973 20d0 6175 ecae

00:00:10;00	942c 942c"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc_str)))
    config = SccWriterConfiguration()
    scc_from_model = scc_writer.from_model(model, config)
    self.assertEqual(scc_from_model, expected_scc)

    # round-trip test
    rt_model = scc_reader.to_model(scc_from_model)
    # cfg = IMSCWriterConfiguration(time_format=TimeExpressionSyntaxEnum.frames, fps=Fraction(30000, 1001))
    # imsc_writer.from_model(rt_model, cfg).write(sys.stdout.buffer)
    b = rt_model.get_body()
    div = list(b)[0]
    p0 = list(div)[0]
    self.assertEqual(Fraction(30 * 1001, 30000), p0.get_begin())
    p1 = list(div)[1]
    self.assertEqual(Fraction(150 * 1001, 30000), p1.get_begin())
    self.assertEqual(Fraction(300 * 1001, 30000), p1.get_end())

  def test_basic_2997NDF(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml">
  <body>
    <div>
      <p begin="3600s" end="3602s">Hello</p>
     </div>
  </body>
</tt>"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc_str)))
    assert model is not None

    expected_scc="""Scenarist_SCC V1.0

00:59:56:03	9420 9420 94ae 94ae 9440 9440 c8e5 ecec ef80 942f 942f

00:59:58:12	942c 942c"""

    config = SccWriterConfiguration(frame_rate=SCCFrameRate.FPS_2997_NDF)
    scc_from_model = scc_writer.from_model(model, config)
    self.assertEqual(scc_from_model, expected_scc)

  def test_basic_30FPS(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:frameRate="30">
  <body>
    <div>
      <p begin="30f" end="90f">Hello</p>
      <p begin="120f" end="150f">Bonjour</p>
     </div>
  </body>
</tt>"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc_str)))
    assert model is not None

    expected_scc="""Scenarist_SCC V1.0

00:00:00:21	9420 9420 94ae 94ae 9440 9440 c8e5 ecec ef80 942f 942f

00:00:03:00	942c 942c

00:00:03:20	9420 9420 94ae 94ae 9440 9440 c2ef 6eea ef75 f280 942f 942f

00:00:05:00	942c 942c"""

    config = SccWriterConfiguration(frame_rate=SCCFrameRate.FPS_30_NDF)
    scc_from_model = scc_writer.from_model(model, config)
    self.assertEqual(scc_from_model, expected_scc)

  def test_zero_start(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:frameRate="30">
  <body>
    <div>
      <p begin="0f" end="90f">Hello</p>
     </div>
  </body>
</tt>"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc_str)))

    with self.assertRaises(RuntimeError):
      scc_writer.from_model(model)

    scc_from_model = scc_writer.from_model(model, SccWriterConfiguration(start_tc="01:00:00;00"))
    expected_scc="""Scenarist_SCC V1.0

00:59:59;21	9420 9420 94ae 94ae 9440 9440 c8e5 ecec ef80 942f 942f

01:00:02;29	942c 942c"""
    self.assertEqual(scc_from_model, expected_scc)

    scc_from_model = scc_writer.from_model(model, SccWriterConfiguration(frame_rate=SCCFrameRate.FPS_30_NDF, start_tc="01:00:00:00"))
    expected_scc="""Scenarist_SCC V1.0

00:59:59:21	9420 9420 94ae 94ae 9440 9440 c8e5 ecec ef80 942f 942f

01:00:03:00	942c 942c"""
    self.assertEqual(scc_from_model, expected_scc)

  def test_multi_regions(self):
    expected_scc="""Scenarist_SCC V1.0

00:00:00;22	9420 9420 94ae 94ae 9440 9440 6180 942f 942f

00:00:01;29	942c 942c

00:00:02;22	9420 9420 94ae 94ae 9440 9440 6280 942f 942f

00:00:03;29	942c 942c"""

    SAMPLE = """<tt xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling" xml:lang="en">
	<head>
		<layout>
			<region xml:id="r_a" tts:origin="0% 0%" tts:extent="50% 50%" />
			<region xml:id="r_b" tts:origin="50% 50%" tts:extent="50% 50%" />
		</layout>
	</head>
	<body>
		<div>
			<p region="r_a" begin="00:00:01.000" end="00:00:02.000">a</p>
			<p region="r_b" begin="00:00:03.000" end="00:00:04.000">b</p>
		</div>
	</body>
</tt>

"""

    model = imsc_reader.to_model(et.ElementTree(et.fromstring(SAMPLE)))
    scc_from_model = scc_writer.from_model(model)
    self.assertEqual(scc_from_model, expected_scc)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the IMSC \\<color\\> parser'''

# pylint: disable=R0201,C0115,C0116

import unittest
from ttconv.utils import parse_color
from ttconv.style_properties import ColorType

class IMSCReaderTest(unittest.TestCase):

  tests = [
    ["#FFFFFF", ColorType((255, 255, 255, 255))],
    ["#FFFFFF7F", ColorType((255, 255, 255, 127))],
    ["rgb(255,128,255)", ColorType((255, 128, 255, 255))],
    ["rgb( 255, 128 ,   255   )", ColorType((255, 128, 255, 255))],
    ["rgba(128,255,255,63)", ColorType((128, 255, 255, 63))],
    ["transparent", ColorType((0, 0, 0, 0))],
    ["black", ColorType((0, 0, 0, 255))],
    ["silver", ColorType((0xc0, 0xc0, 0xc0, 255))],
    ["gray", ColorType((0x80, 0x80, 0x80, 255))],
    ["white", ColorType((255, 255, 255, 255))],
    ["maroon", ColorType((0x80, 0, 0, 255))],
    ["red", ColorType((255, 0, 0, 255))],
    ["purple", ColorType((0x80, 0, 0x80, 255))],
    ["fuchsia", ColorType((255, 0, 255, 255))],
    ["magenta", ColorType((255, 0, 255, 255))],
    ["green", ColorType((0, 0x80, 0, 255))],
    ["lime", ColorType((0, 255, 0, 255))],
    ["olive", ColorType((0x80, 0x80, 0, 255))],
    ["yellow", ColorType((255, 255, 0, 255))],
    ["navy", ColorType((0, 0, 0x80, 255))],
    ["blue", ColorType((0, 0, 255, 255))],
    ["teal", ColorType((0, 0x80, 0x80, 255))],
    ["aqua", ColorType((0, 255, 255, 255))],
    ["cyan", ColorType((0, 255, 255, 255))],
    ["Cyan", ColorType((0, 255, 255, 255))],
    ["CYAN", ColorType((0, 255, 255, 255))],
    ["#FFffFF", ColorType((255, 255, 255, 255))],
    ["#FfFFFF7f", ColorType((255, 255, 255, 127))]
  ]

  def test_bad_color(self):
    with self.assertRaises(ValueError):
      parse_color("#red")

  def test_colors(self):
    for test in self.tests:
      with self.subTest(test[0]):
        c = parse_color(test[0])
        self.assertEqual(c, test[1])

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC caption text"""

from __future__ import annotations

import logging
import copy
from typing import Optional

from ttconv.time_code import SmpteTimeCode

LOGGER = logging.getLogger(__name__)


class SccCaptionText:
  """Caption text content with specific positional, temporal and styling attributes"""

  def __init__(self, text: Optional[str] = ""):
    self._begin: Optional[SmpteTimeCode] = None
    self._end: Optional[SmpteTimeCode] = None
    self._style_properties = {}
    self._text: str = ""
    self._cursor = 0  # Cursor in the text

    if text is not None and text != "":
      self.append(text)

  def set_begin(self, time_code: SmpteTimeCode):
    """Sets begin time code"""
    self._begin = copy.copy(time_code)

  def get_begin(self) -> SmpteTimeCode:
    """Returns the begin time code"""
    return self._begin

  def set_end(self, time_code: SmpteTimeCode):
    """Sets end time code"""
    self._end = copy.copy(time_code)

  def get_end(self) -> SmpteTimeCode:
    """Returns the end time code"""
    return self._end

  def get_text(self) -> str:
    """Returns the text"""
    return self._text

  def get_length(self) -> int:
    """Returns text length"""
    return len(self._text)

  def is_empty(self) -> bool:
    """Returns whether the text is empty or not"""
    return self.get_length() == 0

  def append(self, text: str):
    """Add or replace text content at cursor position"""
    if self._cursor < 0:
      # Insert space characters before current text
      self._text = ' ' * -self._cursor + self._text
      self._cursor = 0

    # print("Append text: ", text, "to", self._text, "at", self._cursor)
    self._text = self._text[:self._cursor] + text + self._text[(self._cursor + len(text)):]
    self._cursor += len(text)
    # print("\t=>", self._text, ", cursor:", self._cursor)

  def set_cursor_at(self, position: int):
    """Set text cursor position"""
    self._cursor = position

  def get_cursor(self) -> int:
    """Returns the cursor position"""
    return self._cursor

  def backspace(self):
    """Remove last character"""
    self._text = self._text[:-1]
    self._cursor = max(self._cursor - 1, 0)

  def get_style_properties(self) -> dict:
    """Sets the style properties map"""
    return self._style_properties

  def add_style_property(self, style_property, value):
    """Adds a style property"""
    if value is None:
      return
    self._style_properties[style_property] = value

  def has_same_style_properties(self, other):
    """Returns whether the current text has the same style properties as the other text"""
    return self._style_properties == other.get_style_properties()

  def __repr__(self):
    return "<" + self.__class__.__name__ + " " + str(self.__dict__) + ">"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2023, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Defines the Least common denominator (LCD) filter."""

from __future__ import annotations
import logging
import typing
from dataclasses import dataclass, field
from numbers import Number

from ttconv.config import ModuleConfiguration
from ttconv.filters.document_filter import DocumentFilter
from ttconv.filters.remove_animations import RemoveAnimationFilter
from ttconv.filters.supported_style_properties import SupportedStylePropertiesFilter
from ttconv.isd import StyleProcessors
from ttconv.model import ContentDocument, ContentElement, Region, P
from ttconv.style_properties import TextAlignType, ColorType, CoordinateType, DisplayAlignType, ExtentType, LengthType, StyleProperties, WritingModeType, NamedColors
import ttconv.utils

LOGGER = logging.getLogger(__name__)

def _replace_regions(element: ContentElement, region_aliases: typing.Mapping[Region, Region]):
  merged_region = region_aliases.get(element.get_region())
  if merged_region is not None:
    element.set_region(merged_region)
  for child in element:
    _replace_regions(child, region_aliases)

def _apply_bg_color(element: ContentElement, bg_color: ColorType):
  if isinstance(element, P):
    element.set_style(StyleProperties.BackgroundColor, bg_color)
  else:
    for child in element:
      _apply_bg_color(child, bg_color)

def _safe_area_decoder(s: Number) -> int:
  safe_area = int(s)
  if 30 < safe_area < 0:
    raise ValueError("Safe area must be an integer between 0 and 30")
  return safe_area

def _color_decoder(s: typing.Optional[ColorType]) -> typing.Optional[ColorType]:
  if s is None:
    return None

  if not isinstance(s, str):
    raise ValueError("Color specification must be a string")

  return ttconv.utils.parse_color(s)

@dataclass
class LCDDocFilterConfig(ModuleConfiguration):
  """Configuration class for the Least common denominator (LCD) filter"""

  @classmethod
  def name(cls):
    return "lcd"

  # specifies the safe area as an integer percentage
  safe_area: typing.Optional[int] = field(default=10, metadata={"decoder": _safe_area_decoder})

  # preserve text alignment
  preserve_text_align: typing.Optional[bool] = field(default=False, metadata={"decoder": bool})

  # overrides the text color
  color: typing.Optional[ColorType] = field(default=None, metadata={"decoder": _color_decoder})

  # overrides the background color
  bg_color: typing.Optional[ColorType] = field(default=None, metadata={"decoder": _color_decoder})

class LCDDocFilter(DocumentFilter):
  """Merges regions and removes all text formatting with the exception of color
  and text alignment."""

  @classmethod
  def get_config_class(cls) -> ModuleConfiguration:
    return LCDDocFilterConfig

  def __init__(self, config: LCDDocFilterConfig):
    super().__init__(config)

  def process(self, doc: ContentDocument) -> ContentDocument:

    # clean-up styles

    supported_styles = {
      StyleProperties.DisplayAlign: [],
      StyleProperties.Extent: [],
      StyleProperties.Origin: [],
      StyleProperties.Position: []
    }

    if self.config.preserve_text_align:
      supported_styles.update({StyleProperties.TextAlign: []})

    if self.config.color is None:
      supported_styles.update({StyleProperties.Color: []})

    if self.config.bg_color is None:
      supported_styles.update({StyleProperties.BackgroundColor: []})

    style_filter = SupportedStylePropertiesFilter(supported_styles)

    style_filter.process_initial_values(doc)

    if doc.get_body() is not None:
      style_filter.process_element(doc.get_body())

    # clean-up animations

    animation_filter = RemoveAnimationFilter()

    if doc.get_body() is not None:
      animation_filter.process_element(doc.get_body())

    # clean-up regions

    initial_extent = doc.get_initial_value(StyleProperties.Extent)
    initial_origin = doc.get_initial_value(StyleProperties.Origin)
    initial_writing_mode = doc.get_initial_value(StyleProperties.WritingMode)
    initial_display_align = doc.get_initial_value(StyleProperties.DisplayAlign)

    retained_regions = dict()
    replaced_regions = dict()

    for region in doc.iter_regions():

      # cleanup animations
      animation_filter.process_element(region)

      # cleanup styles
      style_filter.process_element(region)

      # compute origin
      if (region.get_style(StyleProperties.Origin)) is not None:
        StyleProcessors.Origin.compute(None, region)

      if (region.get_style(StyleProperties.Position)) is not None:
        StyleProcessors.Position.compute(None, region)
        region.set_style(StyleProperties.Position, None)

      if region.get_style(StyleProperties.Origin) is None:
        region.set_style(StyleProperties.Origin, initial_origin if initial_origin is not None \
                         else StyleProperties.Origin.make_initial_value())

      # compute extent
      if (region.get_style(StyleProperties.Extent)) is not None:
        StyleProcessors.Extent.compute(None, region)

      if region.get_style(StyleProperties.Extent) is None:
        region.set_style(StyleProperties.Extent, initial_extent if initial_extent is not None \
                         else StyleProperties.Extent.make_initial_value() )

      # computer writing_mode and display_align

      writing_mode = region.get_style(StyleProperties.WritingMode)
      if writing_mode is None:
        writing_mode = initial_writing_mode if initial_writing_mode is not None \
                        else StyleProperties.WritingMode.make_initial_value()

      display_align = region.get_style(StyleProperties.DisplayAlign)
      if display_align is None:
        display_align = initial_display_align if initial_display_align is not None \
                        else StyleProperties.DisplayAlign.make_initial_value()

      # determine new displayAlign value
      new_display_align = DisplayAlignType.after

      if writing_mode in (WritingModeType.lrtb, WritingModeType.rltb):
        if display_align == DisplayAlignType.before and region.get_style(StyleProperties.Origin).y.value < 50:
          new_display_align = DisplayAlignType.before
        elif region.get_style(StyleProperties.Origin).y.value + region.get_style(StyleProperties.Extent).height.value < 50:
          new_display_align = DisplayAlignType.before
      elif writing_mode == WritingModeType.tblr:
        if display_align == DisplayAlignType.before and region.get_style(StyleProperties.Origin).x.value < 50:
          new_display_align = DisplayAlignType.before
        elif region.get_style(StyleProperties.Origin).x.value + region.get_style(StyleProperties.Extent).width.value < 50:
            new_display_align = DisplayAlignType.before
      else: # writing_mode == WritingModeType.tbrl
        if display_align == DisplayAlignType.before and region.get_style(StyleProperties.Origin).x.value >= 50:
            new_display_align = DisplayAlignType.before
        elif region.get_style(StyleProperties.Origin).x.value + region.get_style(StyleProperties.Extent).width.value >= 50:
            new_display_align = DisplayAlignType.before

      region.set_style(StyleProperties.DisplayAlign, new_display_align)

      # reposition region
      region.set_style(
        StyleProperties.Origin,
        CoordinateType(
          x=LengthType(self.config.safe_area, LengthType.Units.pct),
          y=LengthType(self.config.safe_area, LengthType.Units.pct)
        )
      )

      region.set_style(
        StyleProperties.Extent,
        ExtentType(
          height=LengthType(value=100 - 2 * self.config.safe_area, units=LengthType.Units.pct),
          width=LengthType(value=100 - 2 * self.config.safe_area, units=LengthType.Units.pct)
        )
      )

      # check if a similar region has already been processed

      fingerprint = (
          region.get_begin() or 0,
          region.get_end() or None,
          writing_mode,
          new_display_align
        )

      retained_region = retained_regions.get(fingerprint)

      if retained_region is None:
        retained_regions[fingerprint] = region
      else:
        replaced_regions[region] = retained_region

    # prune aliased regions
    if doc.get_body() is not None:
      _replace_regions(doc.get_body(), replaced_regions)

    for region in list(doc.iter_regions()):
      if region in replaced_regions:
        doc.remove_region(region.get_id())

    # apply background color
    if self.config.bg_color is not None:
      _apply_bg_color(doc.get_body(), self.config.bg_color)

    # apply text color
    if doc.get_body() is not None and self.config.color is not None:
      doc.get_body().set_style(StyleProperties.Color, self.config.color)

    # apply text align
    if doc.get_body() is not None and not self.config.preserve_text_align:
      doc.get_body().set_style(StyleProperties.TextAlign, TextAlignType.center)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the VTT reader"""

# pylint: disable=R0201,C0115,C0116,W0212
import unittest
import io
import os.path

from ttconv.vtt.reader import to_model
import ttconv.style_properties as styles
import ttconv.model as model


class VTTReaderTest(unittest.TestCase):

  def test_sample(self):
    SAMPLE = """WEBVTT

02:00.000 --> 02:05.000
<b>This is bold text</c>

04:00.000 --> 04:05.000
<i>This is italic</i> and this is not
"""
    f = io.StringIO(SAMPLE)
    self.assertIsNotNone(to_model(f))


  def test_samples(self):
    for root, _subdirs, files in os.walk("src/test/resources/vtt/"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".vtt":
          with self.subTest(name):
            with open(os.path.join(root, filename), encoding="utf-8") as f:
              self.assertIsNotNone(to_model(f))

  def test_bold(self):
    f = io.StringIO(r"""WEBVTT

02:00.000 --> 02:05.000
<b>This is bold text</b>
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        break
    else:
      self.fail()

  def test_blank_lines(self):
    # from https://en.wikipedia.org/wiki/SubRip
    SAMPLE = """WEBVTT

1
00:02:16.612 --> 00:02:19.376
Senator, we're making
our final approach into Coruscant.


2
00:02:19.482 --> 00:02:21.609
Very good, Lieutenant.

5
00:03:20.476 --> 00:03:22.671
There was no danger at all.



"""

    f = io.StringIO(SAMPLE)
    self.assertIsNotNone(to_model(f))

  def test_malformed_blank_lines(self):
    # from https://github.com/sandflow/ttconv/issues/439
    # the first cue should be ignored since it is malformed
    SAMPLE = """WEBVTT
Kind: captions
Language: en

00:00:00.799 --> 00:00:02.869 align:start position:0%

hi<00:00:01.040><c> everyone</c><00:00:01.920><c> today</c><00:00:02.240><c> we're</c><00:00:02.399><c> going</c><00:00:02.639><c> to</c><00:00:02.720><c> be</c>

00:00:02.869 --> 00:00:02.879 align:start position:0%
hi everyone today we're going to be
"""

    doc = to_model(io.StringIO(SAMPLE))
    self.assertIsNotNone(doc)
    body = list(doc.get_body())
    self.assertEqual(len(body), 1)
    div = list(body[0])
    self.assertEqual(len(div), 1)

  def test_single_line_with_space(self):
    # from https://github.com/sandflow/ttconv/issues/439
    # the first cue is not ignored since the first line contains a single space
    SAMPLE = """WEBVTT
Kind: captions
Language: en

00:00:00.799 --> 00:00:02.869 align:start position:0%
\x20
hi<00:00:01.040><c> everyone</c><00:00:01.920><c> today</c><00:00:02.240><c> we're</c><00:00:02.399><c> going</c><00:00:02.639><c> to</c><00:00:02.720><c> be</c>

00:00:02.869 --> 00:00:02.879 align:start position:0%
hi everyone today we're going to be
"""

    doc = to_model(io.StringIO(SAMPLE))
    self.assertIsNotNone(doc)
    body = list(doc.get_body())
    self.assertEqual(len(body), 1)
    div = list(body[0])
    self.assertEqual(len(div), 2)

  def test_toplevel_timestamp_tags(self):
    # from https://github.com/sandflow/ttconv/issues/439
    SAMPLE = """WEBVTT
Kind: captions
Language: en

00:00:00.799 --> 00:00:02.869 align:start position:0%
\x20
hi<00:00:01.040><c> everyone</c><00:00:01.920><c> today</c><00:00:02.240><c> we're</c><00:00:02.399><c> going</c><00:00:02.639><c> to</c><00:00:02.720><c> be</c>

00:00:02.869 --> 00:00:02.879 align:start position:0%
hi everyone today we're going to be
"""

    doc = to_model(io.StringIO(SAMPLE))
    body = list(doc.get_body())
    spans_and_brs = list(body[0][0])
    self.assertIsNone(spans_and_brs[0].get_begin()) # \x20
    self.assertIsNone(spans_and_brs[2].get_begin()) # hi
    self.assertEqual(spans_and_brs[3].get_begin(), 1.040 - 0.799) # everyone
    self.assertEqual(spans_and_brs[4].get_begin(), 1.920 - 0.799) # today

  def test_italic(self):
    f = io.StringIO(r"""WEBVTT

00:02:16.612 --> 00:02:19.376
Hello <i>my</i> name is Bob
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        break
    else:
      self.fail()


  def test_underline(self):
    f = io.StringIO(r"""WEBVTT

00:02:16.612 --> 00:02:19.376
Hello <u>my</u> name is Bob
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        break
    else:
      self.fail()


  def test_blue(self):
    f = io.StringIO(r"""WEBVTT

02:00.000 --> 02:05.000
<c.blue>This is bold text</c>
""")

    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.Color) == styles.NamedColors.blue.value:
        break
    else:
      self.fail()


  def test_bg_blue(self):
    f = io.StringIO(r"""WEBVTT

02:00.000 --> 02:05.000
<c.bg_blue>This is bold text</c>
""")

    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.BackgroundColor) == styles.NamedColors.blue.value:
        break
    else:
      self.fail()

  def test_lang(self):
    f = io.StringIO(r"""WEBVTT

02:00.000 --> 02:05.000
<lang es-419>Spanish as used in Latin America and the Caribbean</lang>
""")

    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_lang() == "es-419":
        break
    else:
      self.fail()


  def test_multiline_tags(self):
    f = io.StringIO(r"""WEBVTT

00:02:16.612 --> 00:02:19.376
Hello <b>my
</b> name is Bob
""")
    doc = to_model(f)
    i = doc.get_body().first_child().first_child().dfs_iterator()

    self.assertIsInstance(next(i), model.P)
    self.assertIsInstance(next(i), model.Span)
    self.assertIsInstance(next(i), model.Text)
    e = next(i)
    self.assertIsInstance(e, model.Span)
    self.assertEqual(e.get_style(styles.StyleProperties.FontWeight), styles.FontWeightType.bold)
    self.assertIsInstance(next(i), model.Span)
    self.assertIsInstance(next(i), model.Text)
    self.assertIsInstance(next(i), model.Br)
    self.assertIsInstance(next(i), model.Span)
    self.assertIsInstance(next(i), model.Text)

  def test_long_hours(self):
    f = io.StringIO(r"""WEBVTT

101:00:00.000 --> 101:00:01.000
Hello my name is Bob
""")
    doc = to_model(f)

    self.assertEqual(
      363600,
      doc.get_body().first_child().first_child().get_begin()
    )

    self.assertEqual(
      363601,
      doc.get_body().first_child().first_child().get_end()
    )

  def test_ts_tag(self):
    f = io.StringIO(r"""WEBVTT

00:00:01.000 --> 00:00:03.000
Hello my name<00:02.000>is Bob
""")
    doc = to_model(f)

    self.assertEqual(
      1,
      doc.get_body().first_child().first_child().get_begin()
    )

    second_span = list(doc.get_body().first_child().first_child())[1]

    self.assertEqual(
      1,
      second_span.get_begin()
    )

  def test_ignore_style(self):
    SAMPLE = """WEBVTT

STYLE
::cue { color:lime }

00:00:00.000 --> 00:00:25.000
Red or green?
"""
    f = io.StringIO(SAMPLE)
    self.assertIsNotNone(to_model(f))

  def test_ruby(self):
    # from WPT (bidi_vertical_lr.vrr)
    SAMPLE = """WEBVTT

00:00:00.000 --> 00:00:05.000
<ruby>.<rt>א<c>א</c></rt>ab)<rt>x</rt></ruby>
"""
    f = io.StringIO(SAMPLE)
    self.assertIsNotNone(to_model(f))


  def test_line_origin_extent(self):
    f = io.StringIO(r"""WEBVTT

1
00:00:00.000 --> 00:00:02.000 line:0
Line 0 starting from top

2
00:00:03.000 --> 00:00:05.000 line:5
Line 5 starting from top

3
00:00:06.000 --> 00:00:08.000 line:45%
Line in percentage

4
00:00:09.000 --> 00:00:11.000 line:-1
Line 1 starting from bottom
""")
    doc = to_model(f)
    regions = list(doc.iter_regions())
    # line 0 starting from top
    self.assertEqual(round(regions[0].get_style(styles.StyleProperties.Origin).y.value), 0)
    self.assertEqual(round(regions[0].get_style(styles.StyleProperties.Extent).height.value), 100)
    # line 5 starting from top
    self.assertEqual(round(regions[1].get_style(styles.StyleProperties.Origin).y.value), 31)
    self.assertEqual(round(regions[1].get_style(styles.StyleProperties.Extent).height.value), 69)
    # line in percentage
    self.assertEqual(round(regions[2].get_style(styles.StyleProperties.Origin).y.value), 45)
    self.assertEqual(round(regions[2].get_style(styles.StyleProperties.Extent).height.value), 55)
    # line 1 starting from bottom
    self.assertEqual(round(regions[3].get_style(styles.StyleProperties.Origin).y.value), 94)
    self.assertEqual(round(regions[3].get_style(styles.StyleProperties.Extent).height.value), 6)

  def _cue_settings_to_region(self, settings: str) -> model.Region:
    f = io.StringIO(f"""WEBVTT

1
00:00:00.000 --> 00:00:02.000 {settings}
Line 0 starting from top

""")
    doc = to_model(f)
    regions = list(doc.iter_regions())
    self.assertTrue(len(regions), 1)
    return regions[0]

  def test_pos_20_center(self):
    r = self._cue_settings_to_region("position:20%,center")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.x.value, 0)
    self.assertEqual(e.width.value, 40)

  def test_pos_90_center(self):
    r = self._cue_settings_to_region("position:90%")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.x.value, 80)
    self.assertEqual(e.width.value, 20)

  def test_pos_90_left(self):
    r = self._cue_settings_to_region("position:90%,line-left")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.x.value, 90)
    self.assertEqual(e.width.value, 10)

  def test_pos_90_right(self):
    r = self._cue_settings_to_region("position:90%,line-right")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.x.value, 0)
    self.assertEqual(e.width.value, 90)

  def test_pos_99_left_10(self):
    r = self._cue_settings_to_region("position:99%,line-left size:10")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.x.value, 95)
    self.assertEqual(e.width.value, 5)

  def test_line_99(self):
    r = self._cue_settings_to_region("line:99%")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.y.value, 93.75)
    self.assertEqual(e.height.value, 6.25)

  def test_line_minus_3(self):
    r = self._cue_settings_to_region("line:-3")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.y.value, 81.25)
    self.assertEqual(e.height.value, 18.75)

  def test_line_minus_3_after(self):
    r = self._cue_settings_to_region("line:-3,end")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.y.value, 0)
    self.assertEqual(e.height.value, 81.25)

  def test_line_99_vertical(self):
    r = self._cue_settings_to_region("line:99% vertical:lr")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.x.value, 93.75)
    self.assertEqual(e.width.value, 6.25)

  def test_default_positioning(self):
    r = self._cue_settings_to_region("")
    o : styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    e : styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    self.assertEqual(o.y.value, 100*1/23)
    self.assertEqual(e.height.value, 100 - 2*100*1/23)
    self.assertEqual(o.x.value, 100*1/40)
    self.assertEqual(e.width.value, 100 - 2*100*1/40)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Common utilities'''

import re
import typing
from fractions import Fraction
import ttconv.style_properties as styles

_HEX_COLOR_RE = re.compile(r"#([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})?")
_DEC_COLOR_RE = re.compile(r"rgb\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)")
_DEC_COLORA_RE = re.compile(r"rgba\(\s*(\d+),\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)")

def parse_color(attr_value: str) -> styles.ColorType:
  '''Parses the TTML \\<color\\> value contained in `attr_value`
  '''

  lower_attr_value = str.lower(attr_value)

  if lower_attr_value in styles.NamedColors.__members__:

    return styles.NamedColors[lower_attr_value].value

  m = _HEX_COLOR_RE.match(attr_value)

  if m:

    return styles.ColorType(
      (
        int(m.group(1), 16),
        int(m.group(2), 16),
        int(m.group(3), 16),
        int(m.group(4), 16) if m.group(4) else 255
      )
    )

  m = _DEC_COLOR_RE.match(attr_value)

  if m:

    return styles.ColorType(
      (
        int(m.group(1)),
        int(m.group(2)),
        int(m.group(3)),
        255
      )
    )

  m = _DEC_COLORA_RE.match(attr_value)

  if m:

    return styles.ColorType(
      (
        int(m.group(1)),
        int(m.group(2)),
        int(m.group(3)),
        int(m.group(4))
      )
    )

  raise ValueError("Bad Syntax")


class DisjointIntervals:
  """A set of disjoint intervals"""

  def __init__(self):
    self._intervals: typing.List[typing.Tuple[Fraction, typing.Optional[Fraction]]] = []

  @staticmethod
  def _bisect_left(a, x, lo=0, key=None):
    """Polyfill for bisect.bisect_left with key argument
    TODO: replace with standard library version when Python 3.10+ is minimum requirement
    """
    if lo < 0:
      raise ValueError('lo must be non-negative')
    
    hi = len(a)

    if key is None:
      key = lambda v: v
    
    while lo < hi:
      mid = (lo + hi) // 2
      v = key(a[mid])
      if v is not None and v < x:
        lo = mid + 1
      else:
        hi = mid
    return lo

  @staticmethod
  def _bisect_right(a, x, lo=0, key=None):
    """Polyfill for bisect.bisect_right with key argument
    TODO: replace with standard library version when Python 3.10+ is minimum requirement
    """
    if lo < 0:
      raise ValueError('lo must be non-negative')
    hi = len(a)

    if key is None:
      key = lambda v: v

    while lo < hi:
      mid = (lo + hi) // 2
      v = key(a[mid])
      if v is None or x < v:
        hi = mid
      else:
        lo = mid + 1

    return lo

  @staticmethod
  def _within_interval(x: Fraction, interval: typing.Tuple[Fraction, typing.Optional[Fraction]]) -> bool:
    start, end = interval
    return (x >= start) and (end is None or x < end)

  def add(self, start: Fraction, end: typing.Optional[Fraction]):
    """Adds an interval to the set, merging overlapping intervals"""

    if start is None:
      raise ValueError("Interval start must be specified")

    if end is not None and start >= end:
      raise ValueError("Interval start must be strictly less than end")
    
    if len(self._intervals) == 0:
      self._intervals.append((start, end))
      return

    # look for lower bound
    # find the first interval that ends at or after the new interval starts
    low = DisjointIntervals._bisect_left(self._intervals, start, key=lambda x: x[1])

    if low < len(self._intervals):
      # if such an interval exists, the new start is the minimum of the new start and the start of that interval
      start = min(start, self._intervals[low][0])

    # look for upper bound
    if end is None:
      hi = len(self._intervals)
    else:
      # find the first interval that starts after the new interval ends
      hi = DisjointIntervals._bisect_right(self._intervals, end, lo=low, key=lambda x: x[0])

    if hi > low:
      # if such an interval exists, the new end is the maximum of the new end and the end of the last merged interval
      last_end = self._intervals[hi - 1][1]
      if last_end is None or (end is not None and last_end > end):
        end = last_end

    # replace the range of overlapping intervals with the merged interval
    self._intervals[low : hi] = [(start, end)]

  def contains(self, x: Fraction) -> bool:
    """Returns whether the point is contained in one of the intervals"""
    if len(self._intervals) == 0:
      idx = 0
    if len(self._intervals) == 1:
      idx = 1
    else:
      idx = DisjointIntervals._bisect_right(self._intervals, x, key=lambda i: i[0])

    return idx > 0 and DisjointIntervals._within_interval(x, self._intervals[idx - 1])

  def __len__(self):
    return len(self._intervals)

  def __iter__(self):
    return iter(self._intervals)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the SRT writer"""

# pylint: disable=R0201,C0115,C0116,W0212

import os
import unittest
import xml.etree.ElementTree as et
from fractions import Fraction
from pathlib import Path

import ttconv.imsc.reader as imsc_reader
import ttconv.scc.reader as scc_reader
import ttconv.srt.writer as srt_writer
import ttconv.srt.config as srt_config
from ttconv.model import ContentDocument, Region, Body, Div, P, Span, Text, ContentElement
from ttconv.style_properties import StyleProperties, DisplayType


class SrtWriterTest(unittest.TestCase):

  def test_srt_writer(self):
    doc = ContentDocument()

    r1 = Region("r1", doc)
    doc.put_region(r1)

    r2 = Region("r2", doc)
    r2.set_begin(Fraction(2))
    r2.set_end(Fraction(4))
    doc.put_region(r2)

    body = Body(doc)
    doc.set_body(body)

    div = Div(doc)
    body.push_child(div)

    p = P(doc)
    p.set_region(r1)
    p.set_end(Fraction(2))
    div.push_child(p)

    span = Span(doc)
    span.push_child(Text(doc, "Lorem ipsum dolor sit amet,"))
    p.push_child(span)

    p = P(doc)
    p.set_region(r2)
    div.push_child(p)

    span = Span(doc)
    span.push_child(Text(doc, "consectetur adipiscing elit."))
    p.push_child(span)

    p = P(doc)
    p.set_region(r1)
    p.set_begin(Fraction(4))
    p.set_end(Fraction(6))
    div.push_child(p)

    span = Span(doc)
    span.push_child(Text(doc, "Pellentesque interdum lacinia sollicitudin."))
    p.push_child(span)

    expected_srt = """1
00:00:00,000 --> 00:00:02,000
Lorem ipsum dolor sit amet,

2
00:00:02,000 --> 00:00:04,000
consectetur adipiscing elit.

3
00:00:04,000 --> 00:00:06,000
Pellentesque interdum lacinia sollicitudin.
"""

    srt_from_model = srt_writer.from_model(doc)

    self.assertEqual(expected_srt, srt_from_model)

  def test_scc_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/scc"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".scc":
          with self.subTest(name):
            path = os.path.join(root, filename)
            scc_content = Path(path).read_text()
            test_model = scc_reader.to_model(scc_content)
            srt_from_model = srt_writer.from_model(test_model)
            self.assertTrue(len(srt_from_model) > 0, msg=f"Could not convert {path}")
            self._check_output_srt(test_model, srt_from_model, path)

  def test_imsc_1_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name):
            path = os.path.join(root, filename)
            tree = et.parse(path)
            test_model = imsc_reader.to_model(tree)
            srt_from_model = srt_writer.from_model(test_model)
            self._check_output_srt(test_model, srt_from_model, path)

  @unittest.skip("Too long to process")
  def test_imsc_1_1_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_1/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name):
            path = os.path.join(root, filename)
            tree = et.parse(path)
            test_model = imsc_reader.to_model(tree)
            srt_from_model = srt_writer.from_model(test_model)
            self._check_output_srt(test_model, srt_from_model, path)

  @unittest.skip("IMSC 1.2 is not supported")
  def test_imsc_1_2_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_2/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name):
            path = os.path.join(root, filename)
            tree = et.parse(path)
            test_model = imsc_reader.to_model(tree)
            srt_from_model = srt_writer.from_model(test_model)
            self._check_output_srt(test_model, srt_from_model, path)

  @unittest.skip("IMSC 1.3 is not supported")
  def test_imsc_1_3_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_3/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name):
            path = os.path.join(root, filename)
            tree = et.parse(path)
            test_model = imsc_reader.to_model(tree)
            srt_from_model = srt_writer.from_model(test_model)
            self._check_output_srt(test_model, srt_from_model, path)


  #
  # Utility functions
  #

  def _has_child_paragraphs(self, element: ContentElement) -> bool:
    has_paragraphs = False

    for child in element:
      if child.get_style(StyleProperties.Display) is DisplayType.none and not list(child.iter_animation_steps()):
        continue

      if isinstance(child, P):
        return True

      if isinstance(child, Div):
        has_paragraphs = self._has_child_paragraphs(child)

    return has_paragraphs

  def _has_document_paragraphs(self, doc: ContentDocument) -> bool:
    body = doc.get_body()

    if body is None:
      return False

    paragraphs = False

    for child in body:
      if self._has_child_paragraphs(child):
        return True

    return paragraphs

  def _check_output_srt(self, model: ContentDocument, srt: str, path: str):
    if self._has_document_paragraphs(model):
      self.assertTrue(len(srt) > 0, msg=f"Could not convert {path}")
    else:
      self.assertEqual(0, len(srt), msg=f"Could not convert {path}")

  def test_empty_isds(self):
    tree = et.parse('src/test/resources/ttml/imsc-tests/imsc1/ttml/timing/BasicTiming010.ttml')
    doc = imsc_reader.to_model(tree)
    srt_from_model = srt_writer.from_model(doc)

    self.assertEqual(srt_from_model, """1
00:00:10,000 --> 00:00:24,400
This text must appear at 10 seconds and disappear at 24.4 seconds

2
00:00:25,000 --> 00:00:35,000
This text must appear at 25 seconds and disappear at 35 seconds
""")

  def test_text_formatting_disabled(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en-US" xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling">
  <body>
    <div>
      <p tts:fontStyle="italic" tts:textDecoration="underline" tts:fontWeight="bold" tts:color="red" begin="00:00:00:00" end="00:00:01:00">Lorem</p>
    </div>
  </body>
</tt>"""

    ttml_doc = et.ElementTree(et.fromstring(ttml_doc_str))
    doc = imsc_reader.to_model(ttml_doc)

    config = srt_config.SRTWriterConfiguration.parse({"text_formatting": False})

    srt_from_model = srt_writer.from_model(doc, config)

    self.assertEqual(srt_from_model, """1
00:00:00,000 --> 00:00:01,000
Lorem
""")

  def test_text_italic(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
<tt xml:lang="en-US" xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling">
  <body>
    <div>
      <p tts:fontStyle="italic" begin="00:00:00:00" end="00:00:01:00">Lorem</p>
    </div>
  </body>
</tt>"""

    ttml_doc = et.ElementTree(et.fromstring(ttml_doc_str))
    doc = imsc_reader.to_model(ttml_doc)

    srt_from_model = srt_writer.from_model(doc)

    self.assertEqual(srt_from_model, """1
00:00:00,000 --> 00:00:01,000
<i>Lorem</i>
""")

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2025, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""IMSC 1.1 Text Profile conformance filter.

Validates a ContentDocument against IMSC 1.1 Text Profile constraints
per https://www.w3.org/TR/ttml-imsc1.1/ sections 6, 7.12, and 8.4.

On success, adds the IMSC 1.1 Text Profile designator to the document's
content profiles. On failure, raises ValueError on the first violation found.
"""

from __future__ import annotations

from dataclasses import dataclass

from ttconv.config import ModuleConfiguration
from ttconv.filters.document_filter import DocumentFilter
from ttconv.imsc.designators import IMSC_11_TEXT_PROFILE_DESIGNATOR
from ttconv.isd import ISD
import ttconv.model as model
import ttconv.style_properties as styles

# Section 6: allowed element types.
_ALLOWED_ELEMENT_TYPES = (
  model.Body,
  model.Div,
  model.P,
  model.Span,
  model.Br,
  model.Text,
  model.Ruby,
  model.Rb,
  model.Rt,
  model.Rp,
  model.Rbc,
  model.Rtc,
  model.Region,
)

# Section 6: explicit list of the 36 IMSC 1.1 Text Profile style properties.
_IMSC_11_TEXT_STYLE_PROPS = frozenset([
  styles.StyleProperties.BackgroundColor,
  styles.StyleProperties.Color,
  styles.StyleProperties.Direction,
  styles.StyleProperties.Disparity,
  styles.StyleProperties.Display,
  styles.StyleProperties.DisplayAlign,
  styles.StyleProperties.Extent,
  styles.StyleProperties.FillLineGap,
  styles.StyleProperties.FontFamily,
  styles.StyleProperties.FontSize,
  styles.StyleProperties.FontStyle,
  styles.StyleProperties.FontWeight,
  styles.StyleProperties.LineHeight,
  styles.StyleProperties.LinePadding,
  styles.StyleProperties.LuminanceGain,
  styles.StyleProperties.MultiRowAlign,
  styles.StyleProperties.Opacity,
  styles.StyleProperties.Origin,
  styles.StyleProperties.Overflow,
  styles.StyleProperties.Padding,
  styles.StyleProperties.Position,
  styles.StyleProperties.RubyAlign,
  styles.StyleProperties.RubyPosition,
  styles.StyleProperties.RubyReserve,
  styles.StyleProperties.Shear,
  styles.StyleProperties.ShowBackground,
  styles.StyleProperties.TextAlign,
  styles.StyleProperties.TextCombine,
  styles.StyleProperties.TextDecoration,
  styles.StyleProperties.TextEmphasis,
  styles.StyleProperties.TextOutline,
  styles.StyleProperties.TextShadow,
  styles.StyleProperties.UnicodeBidi,
  styles.StyleProperties.Visibility,
  styles.StyleProperties.WrapOption,
  styles.StyleProperties.WritingMode,
])


def _iter_lengths(value):
  """Yields all LengthType values embedded in a style value."""
  if isinstance(value, styles.LengthType):
    yield value
  elif isinstance(value, styles.ExtentType):
    yield value.width
    yield value.height
  elif isinstance(value, styles.CoordinateType):
    yield value.x
    yield value.y
  elif isinstance(value, styles.PositionType):
    yield value.h_offset
    yield value.v_offset
  elif isinstance(value, styles.PaddingType):
    yield value.before
    yield value.end
    yield value.after
    yield value.start
  elif isinstance(value, styles.TextOutlineType):
    yield value.thickness
  elif isinstance(value, styles.RubyReserveType):
    if value.length is not None:
      yield value.length
  elif isinstance(value, styles.TextShadowType):
    for shadow in value.shadows:
      yield shadow.x_offset
      yield shadow.y_offset
      if shadow.blur_radius is not None:
        yield shadow.blur_radius


def _validate_regions(doc):
  """Validates region-specific structural constraints across the document."""

  has_any_origin = False
  has_any_position = False

  for region in doc.iter_regions():
    # Section 8.4.2: tts:extent SHALL be present on all region elements.
    # Check both specified style and document initial value.
    extent = region.get_style(styles.StyleProperties.Extent)
    if extent is None:
      extent = doc.get_initial_value(styles.StyleProperties.Extent)
    if extent is None:
      raise ValueError(
        f"Region '{region.get_id()}': tts:extent SHALL be present "
        f"on all region elements (section 8.4.2)"
      )

    # Section 8.4.7/8.4.8: origin and position SHALL NOT coexist on same region
    has_origin = region.has_style(styles.StyleProperties.Origin)
    has_position = region.has_style(styles.StyleProperties.Position)
    if has_origin and has_position:
      raise ValueError(
        f"Region '{region.get_id()}': tts:origin and tts:position "
        f"SHALL NOT both be present (section 8.4.7)"
      )

    # Section 8.4.7: tts:origin SHALL use px units or percentage values
    if has_origin:
      has_any_origin = True
      origin = region.get_style(styles.StyleProperties.Origin)
      if origin is not None:
        for length in (origin.x, origin.y):
          if length.units not in (styles.LengthType.Units.pct, styles.LengthType.Units.px):
            raise ValueError(
              f"Region '{region.get_id()}': tts:origin SHALL use px or % "
              f"units only, found '{length.units}' (section 8.4.7)"
            )

    if has_position:
      has_any_position = True

  # Section 8.4.7/8.4.8: document-wide mutual exclusion
  if has_any_origin and has_any_position:
    raise ValueError(
      "tts:origin and tts:position SHALL NOT both be present "
      "in a Document Instance (section 8.4.7/8.4.8)"
    )


def _validate_element_styles(element):
  """Validates specified styles on a ContentDocument element."""

  if not isinstance(element, _ALLOWED_ELEMENT_TYPES):
    raise ValueError(f"Unsupported element type: {type(element).__name__}")

  elem_name = type(element).__name__

  for style_prop in element.iter_styles():
    if style_prop not in _IMSC_11_TEXT_STYLE_PROPS:
      raise ValueError(
        f"{elem_name}: style property "
        f"'{style_prop.__name__}' is not in the IMSC 1.1 Text Profile"
      )

    value = element.get_style(style_prop)

    # Section 8.4.12: linePadding SHALL use c units only
    if style_prop is styles.StyleProperties.LinePadding \
        and value.units != styles.LengthType.Units.c:
      raise ValueError(
        f"{elem_name}: LinePadding SHALL use c units only, "
        f"found '{value.units}' (section 8.4.12)"
      )

    # Section 7.2.18: c units SHALL NOT be present outside of ebutts:linePadding
    if style_prop is not styles.StyleProperties.LinePadding:
      for length in _iter_lengths(value):
        if length.units == styles.LengthType.Units.c:
          raise ValueError(
            f"{elem_name}: {style_prop.__name__} uses 'c' unit, "
            f"only allowed on LinePadding (section 7.2.18)"
          )

    # Section 8.4.13: ebutts:multiRowAlign SHALL only appear on p elements
    if style_prop is styles.StyleProperties.MultiRowAlign \
        and not isinstance(element, model.P):
      raise ValueError(
        f"{elem_name}: MultiRowAlign SHALL only appear on p elements "
        f"(section 8.4.13)"
      )

    # Section 8.4.11: textShadow SHALL NOT have more than 4 shadows
    if style_prop is styles.StyleProperties.TextShadow \
        and value is not styles.SpecialValues.none \
        and len(value.shadows) > 4:
      raise ValueError(
        f"{elem_name}: TextShadow has {len(value.shadows)} shadows, "
        f"maximum 4 allowed (section 8.4.11)"
      )


def _regions_overlap(r1: ISD.Region, r2: ISD.Region) -> bool:
  """Returns True if two ISD regions spatially overlap.

  Compares bounding rectangles defined by tts:origin and tts:extent.
  After ISD computation both properties use rw (x-axis) and rh (y-axis)
  units, so the values are directly comparable.
  """

  # both regions must be "presented", as specified in IMSC, to be considered for overlap
  if not (r1.is_presented() and r2.is_presented()):
    return False

  o1 = r1.get_style(styles.StyleProperties.Origin)
  e1 = r1.get_style(styles.StyleProperties.Extent)
  o2 = r2.get_style(styles.StyleProperties.Origin)
  e2 = r2.get_style(styles.StyleProperties.Extent)

  if o1 is None or e1 is None or o2 is None or e2 is None:
    return False

  # r1 bounding box
  left1 = o1.x.value
  top1 = o1.y.value
  right1 = left1 + e1.width.value
  bottom1 = top1 + e1.height.value

  # r2 bounding box
  left2 = o2.x.value
  top2 = o2.y.value
  right2 = left2 + e2.width.value
  bottom2 = top2 + e2.height.value

  return not (right1 <= left2 or right2 <= left1 or bottom1 <= top2 or bottom2 <= top1)


def _validate_isd_element(element):
  """Validates resolved style values on an ISD element.

  Checks are performed on fully resolved styles (inherited, initial,
  and computed) to catch violations that specified-style checks miss.
  """
  elem_name = type(element).__name__

  # Section 8.4.11: textShadow SHALL NOT have more than 4 shadows
  ts_value = element.get_style(styles.StyleProperties.TextShadow)
  if ts_value is not None and ts_value is not styles.SpecialValues.none:
    if len(ts_value.shadows) > 4:
      raise ValueError(
        f"{elem_name}: TextShadow has {len(ts_value.shadows)} "
        f"shadows, maximum 4 allowed (section 8.4.11)"
      )

  # Section 8.4.9: tts:rubyAlign SHALL be "center" or "spaceAround"
  # (defense-in-depth; model enum already constrains this)
  ruby_align = element.get_style(styles.StyleProperties.RubyAlign)
  if ruby_align is not None and \
      ruby_align not in (styles.RubyAlignType.center, styles.RubyAlignType.spaceAround):
    raise ValueError(
      f"{elem_name}: RubyAlign value '{ruby_align}' not allowed, "
      f"must be center or spaceAround (section 8.4.9)"
    )

  # Section 8.4.10: textOutline thickness SHALL NOT exceed 10% of fontSize
  text_outline = element.get_style(styles.StyleProperties.TextOutline)
  if text_outline is not None and text_outline is not styles.SpecialValues.none:
    font_size = element.get_style(styles.StyleProperties.FontSize)
    if font_size is not None and isinstance(font_size, styles.LengthType) \
        and text_outline.thickness.units == font_size.units:
      if text_outline.thickness.value > 0.1 * font_size.value:
        raise ValueError(
          f"{elem_name}: TextOutline thickness ({text_outline.thickness.value}) "
          f"exceeds 10% of fontSize ({font_size.value}) (section 8.4.10)"
        )




@dataclass
class IMSC11TextFilterConfig(ModuleConfiguration):
  """Configuration for the IMSC 1.1 Text Profile conformance filter."""

  @classmethod
  def name(cls):
    return "imsc11text"


class IMSC11TextFilter(DocumentFilter):
  """Validates a ContentDocument against IMSC 1.1 Text Profile constraints.

  On success, adds the IMSC 1.1 Text Profile designator to the document's
  content profiles. On failure, raises ValueError with the first violation.
  """

  @classmethod
  def get_config_class(cls) -> ModuleConfiguration:
    return IMSC11TextFilterConfig

  def __init__(self, config: IMSC11TextFilterConfig = None):
    super().__init__(config or IMSC11TextFilterConfig())

  def process(self, doc: model.ContentDocument):

    # ContentDocument structural checks on regions
    _validate_regions(doc)

    # ContentDocument specified-style checks on regions
    for region in doc.iter_regions():
      _validate_element_styles(region)

    # ContentDocument specified-style checks on body elements
    body = doc.get_body()
    if body is not None:
      for element in body.dfs_iterator():
        _validate_element_styles(element)

    # Section 7.12.6: px requires tts:extent on tt is not implementable at filter
    # level because the model stores tts:extent on tt as px_resolution (with a
    # default of 1920x1080), making it impossible to distinguish "explicitly set"
    # from "not present". The IMSC reader enforces this at parse time.

    # Section 7.12.7: ttp:frameRate constraints are not implementable at filter
    # level because the model stores timing as Fraction offsets and does not
    # preserve the original format or frameRate attribute.

    # ISD-based resolved style checks
    isd_sequence = ISD.generate_isd_sequence(doc)
    for _, isd in isd_sequence:
      # Section 7.12.1.3: no more than 4 regions per ISD
      regions = list(isd.iter_regions())
      if len(regions) > 4:
        raise ValueError(
          f"ISD has {len(regions)} regions, "
          f"maximum 4 allowed"
        )

      for i in range(0, len(regions)):
        for j in range(i + 1, len(regions)):
          if _regions_overlap(regions[i], regions[j]):
            raise ValueError(
              f"Regions '{regions[i].get_id()}' and '{regions[j].get_id()}' "
              f"spatially overlap"
            )

      for isd_region in regions:
        for element in isd_region.dfs_iterator():
          _validate_isd_element(element)

    # Success: add designator to content profiles
    content_profiles = doc.get_content_profiles()
    if content_profiles is None:
      content_profiles = set()
    content_profiles.add(IMSC_11_TEXT_PROFILE_DESIGNATOR)
    doc.set_content_profiles(content_profiles)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC Special characters"""

from __future__ import annotations

import typing

from ttconv.scc.codes import SccCode


class SccSpecialCharacter(SccCode):
  """SCC Special character definition"""

  # Special characters specified in CEA-608
  REGISTERED_MARK_SYMBOL = (0x1130, 0x1930, '\u00AE')  # ® Registered mark symbol
  DEGREE_SIGN = (0x1131, 0x1931, '\u00B0')  # ° Degree sign
  VULGAR_FRACTION_ONE_HALF = (0x1132, 0x1932, '\u00BD')  # ¿ Inverse question mark
  INVERTED_QUESTION_MARK = (0x1133, 0x1933, '\u00BF')  # ½ Vulgar one half fraction
  TRADEMARK_SYMBOL = (0x1134, 0x1934, '\u2122')  # ™ Trademark symbol
  CENTS_SIGN = (0x1135, 0x1935, '\u00A2')  # ¢ Cents sign
  POUNDS_STERLING_SIGN = (0x1136, 0x1936, '\u00A3')  # £ Pounds Sterling sign
  MUSIC_NOTE = (0x1137, 0x1937, '\u266A')  # ♪ Music note
  LOWER_CASE_A_WITH_GRAVE_ACCENT = (0x1138, 0x1938, '\u00E0')  # à Lower-case a with grave accent
  TRANSPARENT_SPACE = (0x1139, 0x1939, '\u0020')  # Transparent space
  LOWER_CASE_E_WITH_GRAVE_ACCENT = (0x113A, 0x193A, '\u00E8')  # è Lower-case e with grave accent
  LOWER_CASE_A_WITH_CIRCUMFLEX = (0x113B, 0x193B, '\u00E2')  # â Lower-case a with circumflex
  LOWER_CASE_E_WITH_CIRCUMFLEX = (0x113C, 0x193C, '\u00EA')  # ê Lower-case e with circumflex
  LOWER_CASE_I_WITH_CIRCUMFLEX = (0x113D, 0x193D, '\u00EE')  # î Lower-case i with circumflex
  LOWER_CASE_O_WITH_CIRCUMFLEX = (0x113E, 0x193E, '\u00F4')  # ô Lower-case o with circumflex
  LOWER_CASE_U_WITH_CIRCUMFLEX = (0x113F, 0x193F, '\u00FB')  # û Lower-case u with circumflex

  def __init__(self, channel_1: int, channel_2: int, unicode: chr):
    super().__init__(channel_1, channel_2)
    self._unicode = unicode

  def get_unicode_value(self) -> chr:
    """Returns the special or extended character unicode value"""
    return self._unicode

  @staticmethod
  def from_unicode(c: str) -> typing.Optional[SccSpecialCharacter]:
    """Find the special character corresponding to the Unicode character"""
    for spec_char in list(SccSpecialCharacter):
      if spec_char.get_unicode_value() == c:
        return spec_char
    return None

  @staticmethod
  def find(value: int) -> typing.Optional[SccSpecialCharacter]:
    """Find the special character corresponding to the specified value"""
    for spec_char in list(SccSpecialCharacter):
      if spec_char.contains_value(value):
        return spec_char
    return None



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC line"""

from __future__ import annotations

import logging
import re
from typing import List, Optional

from ttconv.scc.caption_style import SccCaptionStyle
from ttconv.scc.codes import SccChannel
from ttconv.scc.codes.attribute_codes import SccAttributeCode
from ttconv.scc.codes.control_codes import SccControlCode
from ttconv.scc.codes.extended_characters import SccExtendedCharacter
from ttconv.scc.codes.mid_row_codes import SccMidRowCode
from ttconv.scc.codes.preambles_address_codes import SccPreambleAddressCode
from ttconv.scc.codes.special_characters import SccSpecialCharacter
from ttconv.scc.context import SccContext
from ttconv.scc.disassembly import get_scc_word_disassembly
from ttconv.scc.word import SccWord
from ttconv.time_code import SmpteTimeCode, FPS_30

LOGGER = logging.getLogger(__name__)

NDF_PATTERN = ':'.join(['(?P<ndf_h>[0-9]{2})',
                      '(?P<ndf_m>[0-9]{2})',
                      '(?P<ndf_s>[0-9]{2})',
                      '(?P<ndf_f>[0-9]{2})'])
DF_PATTERN = '(:|;|.|,)'.join(['(?P<df_h>[0-9]{2})',
                              '(?P<df_m>[0-9]{2})',
                              '(?P<df_s>[0-9]{2})',
                              '(?P<df_f>[0-9]{2})'])
SCC_LINE_PATTERN = '((' + NDF_PATTERN + ')|(' + DF_PATTERN + '))\t.*'


class SccLine:
  """SCC line definition"""

  def __init__(self, time_code: SmpteTimeCode, scc_words: List[SccWord]):
    self.time_code = time_code
    self.scc_words = scc_words

  @staticmethod
  def from_str(line: str) -> Optional[SccLine]:
    """Creates a SCC line instance from the specified string"""
    if not line:
      return None

    regex = re.compile(SCC_LINE_PATTERN)
    match = regex.match(line)

    if match is None:
      return None

    time_code = match.group(1)
    time_offset = SmpteTimeCode.parse(time_code, FPS_30)

    hex_words = line.split('\t')[1].split(' ')
    scc_words = [SccWord.from_str(hex_word) for hex_word in hex_words if hex_word]
    return SccLine(time_offset, scc_words)

  def get_style(self) -> SccCaptionStyle:
    """Analyses the line words to find SCC control codes and define the caption style"""
    scc_words = self.scc_words
    if not scc_words:
      return SccCaptionStyle.Unknown

    for word in scc_words:
      prefix = SccControlCode.find(word.value)

      if prefix in [SccControlCode.RU2, SccControlCode.RU3, SccControlCode.RU4]:
        return SccCaptionStyle.RollUp

      if prefix is SccControlCode.RDC:
        return SccCaptionStyle.PaintOn

      if prefix is SccControlCode.RCL:
        return SccCaptionStyle.PopOn

    return SccCaptionStyle.Unknown

  def to_disassembly(self, show_channels = False) -> str:
    """Converts SCC line into the disassembly format"""
    disassembly_line = str(self.time_code) + "\t"

    for scc_word in self.scc_words:
      disassembly_line += get_scc_word_disassembly(scc_word, show_channels)

    return disassembly_line

  def process(self, context: SccContext) -> SmpteTimeCode:
    """Converts the SCC line to the data model"""

    debug = str(self.time_code) + "\t"

    first = True
    for scc_word in self.scc_words:
      if not first:
        self.time_code.add_frames()
      first = False

      if context.previous_word is not None and context.previous_word.value == scc_word.value and context.previous_word.is_code():
        context.previous_word = None
        continue

      if scc_word.value == 0x0000:
        continue

      if scc_word.byte_1 < 0x20:

        scc_code = scc_word.get_code()
        caption_channel = scc_word.get_channel()

        if caption_channel is not SccChannel.CHANNEL_1:
          if context.current_channel is not caption_channel:
            LOGGER.warning("Skip Caption Channel 2 content")
          context.current_channel = caption_channel
          continue

        context.current_channel = caption_channel

        if isinstance(scc_code, SccPreambleAddressCode):
          debug += scc_code.debug(scc_word.value)
          context.process_preamble_address_code(scc_code, self.time_code)
          context.previous_word_type = type(scc_code)

        elif isinstance(scc_code, SccAttributeCode):
          debug += scc_code.debug(scc_word.value)
          context.process_attribute_code(scc_code, self.time_code)
          context.previous_word_type = type(scc_code)

        elif isinstance(scc_code, SccMidRowCode):
          debug += scc_code.debug(scc_word.value)
          context.process_mid_row_code(scc_code, self.time_code)
          context.previous_word_type = type(scc_code)

        elif isinstance(scc_code, SccControlCode):
          debug += scc_code.debug(scc_word.value)
          context.process_control_code(scc_code, self.time_code)
          context.previous_word_type = type(scc_code)

        elif isinstance(scc_code, SccSpecialCharacter):
          word = scc_code.get_unicode_value()
          debug += word
          context.process_text(word, self.time_code)
          context.previous_word_type = type(scc_code)

        elif isinstance(scc_code, SccExtendedCharacter):
          context.backspace(self.time_code)

          word = scc_code.get_unicode_value()
          debug += word
          context.process_text(word, self.time_code)
          context.previous_word_type = type(scc_code)

        else:
          debug += "[??/" + hex(scc_word.value) + "]"
          LOGGER.warning("Unsupported SCC word: %s", hex(scc_word.value))
          context.previous_word_type = None

      else:
        if context.current_channel is not SccChannel.CHANNEL_1:
          # LOGGER.warning("Skip Caption Channel 2 code")
          continue

        text = scc_word.to_text()
        debug += text
        context.process_text(text, self.time_code)
        context.previous_word_type = str

      context.previous_word = scc_word

    LOGGER.debug(debug)

    return self.time_code



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2021, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the VTT style"""

# pylint: disable=R0201,C0115,C0116,W0212

import unittest

from ttconv.model import Span
from ttconv.vtt import style
from ttconv.style_properties import StyleProperties, FontWeightType, FontStyleType, TextDecorationType, ColorType, NamedColors


class VttStyleTest(unittest.TestCase):

  def test_style_is_bold(self):
    element = Span()

    self.assertFalse(style.is_element_bold(element))

    element.set_style(StyleProperties.FontWeight, FontWeightType.normal)

    self.assertFalse(style.is_element_bold(element))

    element.set_style(StyleProperties.FontWeight, FontWeightType.bold)

    self.assertTrue(style.is_element_bold(element))

  def test_style_is_italic(self):
    element = Span()

    self.assertFalse(style.is_element_italic(element))

    element.set_style(StyleProperties.FontStyle, FontStyleType.oblique)

    self.assertFalse(style.is_element_italic(element))

    element.set_style(StyleProperties.FontStyle, FontStyleType.normal)

    self.assertFalse(style.is_element_italic(element))

    element.set_style(StyleProperties.FontStyle, FontStyleType.italic)

    self.assertTrue(style.is_element_italic(element))

  def test_style_is_underlined(self):
    element = Span()

    self.assertFalse(style.is_element_underlined(element))

    element.set_style(StyleProperties.TextDecoration, TextDecorationType(line_through=True))

    self.assertFalse(style.is_element_underlined(element))

    element.set_style(StyleProperties.TextDecoration, TextDecorationType(overline=True))

    self.assertFalse(style.is_element_underlined(element))

    element.set_style(StyleProperties.TextDecoration, TextDecorationType(underline=True))

    self.assertTrue(style.is_element_underlined(element))

  def test_style_color(self):
    element = Span()

    self.assertIsNone(style.get_color(element))

    element.set_style(StyleProperties.Color, None)

    self.assertIsNone(style.get_color(element))

    element.set_style(StyleProperties.Color, ColorType((0x12, 0x34, 0x56, 0x78)))

    self.assertEqual("#12345678", style.get_color(element))

    element.set_style(StyleProperties.Color, NamedColors.white.value)

    self.assertEqual("#ffffffff", style.get_color(element))

    element.set_style(StyleProperties.Color, NamedColors.transparent.value)

    self.assertEqual("#00000000", style.get_color(element))

  def test_style_background_color(self):
    element = Span()

    self.assertIsNone(style.get_background_color(element))

    element.set_style(StyleProperties.BackgroundColor, None)

    self.assertIsNone(style.get_background_color(element))

    element.set_style(StyleProperties.BackgroundColor, ColorType((0x12, 0x34, 0x56, 0x78)))

    self.assertEqual("#12345678", style.get_background_color(element))

    element.set_style(StyleProperties.BackgroundColor, NamedColors.white.value)

    self.assertEqual("#ffffffff", style.get_background_color(element))

    element.set_style(StyleProperties.BackgroundColor, NamedColors.transparent.value)

    self.assertEqual("#00000000", style.get_background_color(element))

  def test_style_color_classname(self):
    self.assertEqual("fg_color_00000000", style.get_color_classname("#00000000"))

    self.assertEqual("white", style.get_color_classname("#ffffffff"))

    self.assertEqual("lime", style.get_color_classname("#00ff00ff"))

    self.assertEqual("cyan", style.get_color_classname("#00ffffff"))

    self.assertEqual("red", style.get_color_classname("#ff0000ff"))

    self.assertEqual("yellow", style.get_color_classname("#ffff00ff"))

    self.assertEqual("magenta", style.get_color_classname("#ff00ffff"))

    self.assertEqual("blue", style.get_color_classname("#0000ffff"))

    self.assertEqual("black", style.get_color_classname("#000000ff"))

  def test_style_background_color_classname(self):
    self.assertEqual("bg_color_00000000", style.get_background_color_classname("#00000000"))

    self.assertEqual("bg_white", style.get_background_color_classname("#ffffffff"))

    self.assertEqual("bg_lime", style.get_background_color_classname("#00ff00ff"))

    self.assertEqual("bg_cyan", style.get_background_color_classname("#00ffffff"))

    self.assertEqual("bg_red", style.get_background_color_classname("#ff0000ff"))

    self.assertEqual("bg_yellow", style.get_background_color_classname("#ffff00ff"))

    self.assertEqual("bg_magenta", style.get_background_color_classname("#ff00ffff"))

    self.assertEqual("bg_blue", style.get_background_color_classname("#0000ffff"))

    self.assertEqual("bg_black", style.get_background_color_classname("#000000ff"))

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the IMSC \\<time-expression\\> parser'''

# pylint: disable=R0201,C0115,C0116

import unittest
from fractions import Fraction

from ttconv.imsc.attributes import DropMode, TemporalAttributeParsingContext, TimeBase, parse_time_expression
from ttconv.time_code import SmpteTimeCode

class IMSCTimeExpressionsTest(unittest.TestCase):

  tests = [
    ("1.2s", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(12, 10)),
    ("1.2m", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(72)),
    ("1.2h", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(4320)),
    ("24f", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(1001, 1000)),
    ("120t", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(2)),
    ("01:02:03", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(3723)),
    ("01:02:03.235", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(3723235, 1000)),
    ("01:02:03.2350", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(3723235, 1000)),
    ("01:02:03:20", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(3723) + 20/Fraction(24000, 1001)),
    ("100:00:00.1", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), 360000 + Fraction(1, 10)),
    ("100:00:00:10", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), 360000 + 10/Fraction(24000, 1001)),
    ("00:00:15:00", TemporalAttributeParsingContext(Fraction(24000, 1001), 60), Fraction(15, 1)),
    (
      "01:02:03:20",
      TemporalAttributeParsingContext(Fraction(30000, 1001), 60, TimeBase.smpte),
      (3723 * 30 + 20)/Fraction(30000, 1001)
    ),
    (
      "01:02:03:20",
      TemporalAttributeParsingContext(Fraction(30000, 1001), 60, TimeBase.smpte, DropMode.dropNTSC),
      SmpteTimeCode(1, 2, 3, 20, Fraction(30000, 1001), True).to_temporal_offset()
    ),
  ]

  def test_timing_expressions(self):
    for test in self.tests:
      with self.subTest(test[0]):
        c = parse_time_expression(test[1], test[0])
        self.assertEqual(c, test[2])
        self.assertTrue(isinstance(c, Fraction))

  def test_bad_frame_count(self):
    with self.assertRaises(ValueError):
      parse_time_expression(TemporalAttributeParsingContext(), "100:00:00:100")

  def test_bad_frame_count_non_strict(self):
    self.assertEqual(
      parse_time_expression(TemporalAttributeParsingContext(Fraction(24)), "100:00:00:100", False),
        360000 + Fraction(23, 24)
    )

  def test_bad_syntax(self):
    with self.assertRaises(ValueError):
      parse_time_expression(TemporalAttributeParsingContext(), "100:00:00;01")


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SRT style"""

from typing import Optional

from ttconv.model import ContentElement
from ttconv.style_properties import FontWeightType, StyleProperties, FontStyleType, TextDecorationType

BOLD_TAG_IN = "<b>"
BOLD_TAG_OUT = "</b>"

ITALIC_TAG_IN = "<i>"
ITALIC_TAG_OUT = "</i>"

UNDERLINE_TAG_IN = "<u>"
UNDERLINE_TAG_OUT = "</u>"

FONT_COLOR_TAG_IN = "<font color=\"{}\">"
FONT_COLOR_TAG_OUT = "</font>"


def is_element_bold(element: ContentElement) -> bool:
  """Returns whether the element text is bold"""
  font_weight: Optional[FontWeightType] = element.get_style(StyleProperties.FontWeight)
  return font_weight is not None and font_weight is FontWeightType.bold


def is_element_italic(element: ContentElement) -> bool:
  """Returns whether the element text is italic"""
  font_style: Optional[FontStyleType] = element.get_style(StyleProperties.FontStyle)
  return font_style is not None and font_style is FontStyleType.italic


def is_element_underlined(element: ContentElement) -> bool:
  """Returns whether the element text is underlined"""
  text_decoration: Optional[TextDecorationType] = element.get_style(StyleProperties.TextDecoration)
  return text_decoration is not None and text_decoration.underline is True


def get_font_color(element: ContentElement) -> Optional[str]:
  """Returns the font color code if present"""
  font_color = element.get_style(StyleProperties.Color)

  if font_color is None:
    return None

  (r, g, b, a) = font_color.components
  return "#{:02x}{:02x}{:02x}{:02x}".format(r, g, b, a)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the Time Code"""

# pylint: disable=R0201,C0115,C0116,W0212

from fractions import Fraction
import unittest

from ttconv.time_code import ClockTime


class TimeCodeTest(unittest.TestCase):

  def test_time_code(self):
    seconds = Fraction(2469, 20)
    time_code = ClockTime.from_seconds(seconds)
    self.assertEqual("00:02:03.450", str(time_code))
    time_code.set_separator(",")
    self.assertEqual("00:02:03,450", str(time_code))
    self.assertEqual(seconds, time_code.to_seconds())

    seconds = 3723.4567
    time_code = ClockTime.from_seconds(seconds)
    self.assertEqual("01:02:03.457", str(time_code))
    time_code.set_separator(",")
    self.assertEqual("01:02:03,457", str(time_code))
    self.assertAlmostEqual(seconds, time_code.to_seconds(), delta=0.001)

    seconds = Fraction(3600, 4)
    time_code = ClockTime.from_seconds(seconds)
    self.assertEqual("00:15:00.000", str(time_code))
    time_code.set_separator(",")
    self.assertEqual("00:15:00,000", str(time_code))
    self.assertEqual(seconds, time_code.to_seconds())

    one_day_in_seconds = 24 * 3600

    seconds = one_day_in_seconds + 1.500
    time_code = ClockTime.from_seconds(seconds)
    self.assertEqual("24:00:01.500", str(time_code))
    time_code.set_separator(",")
    self.assertEqual("24:00:01,500", str(time_code))
    self.assertEqual(seconds, time_code.to_seconds())

  def test_from_seconds_rounding_up(self):
      seconds = Fraction(41039999,7500)
      time_code = ClockTime.from_seconds(seconds)
      self.assertEqual(time_code.get_hours(), 1)
      self.assertEqual(time_code.get_minutes(), 31)
      self.assertEqual(time_code.get_seconds(), 12)
      self.assertEqual(time_code.get_milliseconds(), 0)


  def test_from_seconds_rounding_down(self):
      seconds = 5471.9994
      time_code = ClockTime.from_seconds(seconds)
      self.assertEqual(time_code.get_hours(), 1)
      self.assertEqual(time_code.get_minutes(), 31)
      self.assertEqual(time_code.get_seconds(), 11)
      self.assertEqual(time_code.get_milliseconds(), 999)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2021, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Processes an EBU STL data file into the canonical model"""

from numbers import Number
import logging
import typing
import collections
import struct
from fractions import Fraction
import sys

from ttconv import model
from ttconv.imsc.designators import IMSC_11_TEXT_PROFILE_DESIGNATOR
import ttconv.style_properties as styles
from ttconv.stl import tf
from ttconv.time_code import SmpteTimeCode

LOGGER = logging.getLogger(__name__)

DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT = 5
DEFAULT_VERTICAL_SAFE_MARGIN_PCT = 10
DEFAULT_TELETEXT_ROWS = 23
DEFAULT_TELETEXT_COLS = 40
DEFAULT_LINE_HEIGHT_PCT = 125
DEFAULT_SINGLE_HEIGHT_FONT_SIZE_PCT = 80
DEFAULT_DOUBLE_HEIGHT_FONT_SIZE_PCT = 160
LINE_PADDING_LENGTH_C = 0.5
DEFAULT_FONT_STACK = ("Verdana", "Arial", "Tiresias", styles.GenericFontFamilyType.sansSerif)

_LC_BCP47_MAP = {
  b'00': "",
  b'01': "sq",
  b'02': "br",
  b'03': "ca",
  b'04': "hr",
  b'05': "cy",
  b'06': "cs",
  b'07': "da",
  b'08': "de",
  b'09': "en",
  b'0A': "es",
  b'0B': "eo",
  b'0C': "et",
  b'0D': "eu",
  b'0E': "fo",
  b'0F': "fr",
  b'10': "fy",
  b'11': "ga",
  b'12': "gd",
  b'13': "gl",
  b'14': "is",
  b'15': "it",
  b'16': "se",
  b'17': "la",
  b'18': "lv",
  b'19': "lb",
  b'1A': "lt",
  b'1B': "hu",
  b'1C': "mt",
  b'1D': "nl",
  b'1E': "no",
  b'1F': "oc",
  b'20': "pl",
  b'21': "pt",
  b'22': "ro",
  b'23': "rm",
  b'24': "sr",
  b'25': "sk",
  b'26': "sl",
  b'27': "fi",
  b'28': "sv",
  b'29': "tr",
  b'2A': "nl",
  b'2B': "wa",
  # b'2C'-b'3F' have no mappings
  b'7F': "am",
  b'7E': "ar",
  b'7D': "hy",
  b'7C': "as",
  b'7B': "az",
  b'7A': "bm",
  b'79': "be",
  b'78': "bn",
  b'77': "bg",
  b'76': "my",
  b'75': "zh",
  b'74': "cv",
  b'73': "prs",
  b'72': "ff",
  b'71': "ka",
  b'70': "el",
  b'6F': "gu",
  b'6E': "gn",
  b'6D': "ha",
  b'6C': "he",
  b'6B': "hi",
  b'6A': "id",
  b'69': "ja",
  b'68': "kn",
  b'67': "kk",
  b'66': "km",
  b'65': "ko",
  b'64': "lo",
  b'63': "mk",
  b'62': "mg",
  b'61': "ms",
  b'60': "mo",
  b'5F': "mr",
  b'5E': "nd",
  b'5D': "ne",
  b'5C': "or",
  b'5B': "pap",
  b'5A': "fa",
  b'59': "pa",
  b'58': "ps",
  b'57': "qu",
  b'56': "ru",
  # b'55' (Ruthenian) has no mappings
  b'54': "sh",
  b'53': "sn",
  b'52': "si",
  b'51': "so",
  b'50': "srn",
  b'4F': "sw",
  b'4E': "tg",
  b'4D': "ta",
  b'4C': "tt",
  b'4B': "te",
  b'4A': "th",
  b'49': "uk",
  b'48': "ur",
  b'47': "uz",
  b'46': "vi",
  b'45': "zu",
  # b'44'-b'40' are undefined
}

_DFC_FRACTION_MAP = {
  b'STL23.01': Fraction(24000, 1001),
  b'STL24.01': Fraction(24),
  b'STL25.01': Fraction(25),
  b'STL30.01': Fraction(30000, 1001),
  b'STL50.01': Fraction(50)
}

_GSIBlock = collections.namedtuple(
    "GSI",
    ["CPN", "DFC", "DSC", "CCT", "LC", "OPT", "OET", "TPT", "TET", "TN", "TCD", "SLR", "CD", "RD", "RN", \
      "TNB", "TNS", "TNG", "MNC", "MNR", "TCS", "TCP", "TCF", "TND", "DSN", "CO", "PUB", "EN", "ECD", "UDA"]
  )

_TTIBlock = collections.namedtuple(
    "TTI", ["SGN", "SN", "EBN", "CS", "TCIh", "TCIm", "TCIs", "TCIf", "TCOh", "TCOm", "TCOs", "TCOf", "VP", "JC", "CF", "TF"]
  )

def _get_region_from_model(
  doc: model.ContentDocument,
  x_origin: Number,
  y_origin: Number,
  width: Number,
  height: Number,
  display_align: styles.DisplayAlignType
  ):
  """Returns a matching region from `doc` or creates one
  """

  found_region = None

  regions = list(doc.iter_regions())

  for r in regions:
    r_origin: styles.CoordinateType = r.get_style(styles.StyleProperties.Origin)
    assert r_origin is not None
    assert r_origin.x.units is styles.LengthType.Units.pct
    assert r_origin.y.units is styles.LengthType.Units.pct
    if r_origin.x.value != x_origin or r_origin.y.value != y_origin:
      continue

    r_extent: styles.ExtentType = r.get_style(styles.StyleProperties.Extent)
    assert r_extent is not None
    assert r_extent.height.units is styles.LengthType.Units.pct
    assert r_extent.width.units is styles.LengthType.Units.pct
    if r_extent.height.value != height or r_extent.width.value != width:
      continue  

    r_display_align: styles.DisplayAlignType = r.get_style(styles.StyleProperties.DisplayAlign)
    assert r_display_align is not None
    if r_display_align != display_align:
      continue

    found_region = r
    break

  if found_region is None:
    found_region = model.Region(f"r{len(regions)}", doc)
    found_region.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        height=styles.LengthType(height, styles.LengthType.Units.pct),
        width=styles.LengthType(width, styles.LengthType.Units.pct),
      )
      )
    found_region.set_style(
      styles.StyleProperties.Origin,
      styles.CoordinateType(
        x=styles.LengthType(x_origin, styles.LengthType.Units.pct),
        y=styles.LengthType(y_origin, styles.LengthType.Units.pct)
      )
    )
    found_region.set_style(
      styles.StyleProperties.DisplayAlign,
      display_align
    )
    doc.put_region(found_region)
  
  return found_region

class DataFile:
  """Represents an EBU STL datafile
  """

  def __init__(
    self,
    gsi_block: bytes,
    disable_fill_line_gap: bool = False,
    disable_line_padding: bool = False,
    start_tc: typing.Optional[str] = None,
    font_stack: typing.Tuple[typing.Union[str, styles.GenericFontFamilyType]] = None,
    max_row_count: typing.Optional[typing.Union[int, str]] = None
    ):
    
    self.gsi = _GSIBlock._make(
      struct.unpack(
        '3s8sc2s2s32s32s32s32s32s32s16s6s6s2s5s5s3s2s2s1s8s8s1s1s3s32s32s32s75x576s', gsi_block
      )
    )

    self.doc = model.ContentDocument()

    # we are guaranteeing that the result conforms with IMSC 1.1 Text Profile
    self.doc.set_content_profiles(set([IMSC_11_TEXT_PROFILE_DESIGNATOR]))

    self.doc.set_cell_resolution(
      model.CellResolutionType(
        columns=round(100 * DEFAULT_TELETEXT_COLS / (100 - 2 * DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT)),
        rows=round(100 * DEFAULT_TELETEXT_ROWS / (100 - 2 * DEFAULT_VERTICAL_SAFE_MARGIN_PCT))
      )
    )

    self.doc.set_active_area(
      model.ActiveAreaType(
        left_offset=DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT / 100,
        top_offset=DEFAULT_VERTICAL_SAFE_MARGIN_PCT / 100,
        width=1 - 2 * DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT / 100,
        height=1 - 2 * DEFAULT_VERTICAL_SAFE_MARGIN_PCT / 100
      )
    )

    self.body = model.Body(self.doc)
    
    if not disable_fill_line_gap:
      self.body.set_style(
        styles.StyleProperties.FillLineGap,
        True
      )

    if not disable_line_padding:
      self.body.set_style(
        styles.StyleProperties.LinePadding,
        styles.LengthType(
          LINE_PADDING_LENGTH_C,
          styles.LengthType.Units.c
        )
      )

    if font_stack is not None:
      self.body.set_style(
        styles.StyleProperties.FontFamily,
        font_stack
      )
    else:
      self.body.set_style(
        styles.StyleProperties.FontFamily,
        DEFAULT_FONT_STACK
      )

    self.doc.set_body(self.body)

    self.sgn_to_div_map = {}

    self.last_sn = None

    self.is_in_extension = False

    self.tti_tf = None

    self.fps = _DFC_FRACTION_MAP.get(self.gsi.DFC)
    if self.fps is None:
      LOGGER.error("Unknown GSI DFC value %s, defaulting to 25 fps", self.gsi.DFC)
      self.fps = Fraction(25)
    else:
      LOGGER.debug("GSI DFC: %s", self.gsi.DFC)

    self.cct = self.gsi.CCT
    LOGGER.debug("GSI CCT: %s", self.gsi.CCT)

    try:
      self.tti_count = int(self.gsi.TNB)
      LOGGER.debug("GSI TNB: %s", self.gsi.TNB)
    except ValueError:
      LOGGER.error("Invalid TNB field value: %s", self.gsi.TNB)
      self.tti_count = sys.maxsize

    self.language = _LC_BCP47_MAP.get(self.gsi.LC)
    if self.language is None:
      LOGGER.warning("Unknown LC value: %s, defaulting to 'unspecified''", self.gsi.LC)
      self.language = ""
    else:
      LOGGER.debug("GSI LC: %s", self.gsi.LC)

    self.doc.set_lang(self.language)

    if start_tc is None:
      self.start_offset = 0
    elif start_tc == "TCP":
      try:
        self.start_offset = SmpteTimeCode(
            int(self.gsi.TCP[0:2]),
            int(self.gsi.TCP[2:4]),
            int(self.gsi.TCP[4:6]),
            int(self.gsi.TCP[6:8]),
            self.get_fps()
          ).to_temporal_offset()
        LOGGER.debug("GSI TCP: %s", self.gsi.TCP)
      except ValueError:
        LOGGER.error("Invalid TCP value: %s", self.gsi.tcp)
        self.start_offset = 0
    else:
      try:
        self.start_offset = SmpteTimeCode.parse(start_tc, self.get_fps()).to_temporal_offset()
      except ValueError:
        LOGGER.error("Invalid start_tc value")
        raise

    LOGGER.debug("GSI DSC: %s", self.gsi.DSC)
    if max_row_count is None or self.is_teletext():
      self.max_row_count = DEFAULT_TELETEXT_ROWS
    elif isinstance(max_row_count, str) and max_row_count == "MNR":
      try:
        self.max_row_count = int(self.gsi.MNR)
        LOGGER.debug("GSI MNR: %s", self.gsi.MNR)
      except ValueError:
        LOGGER.error("Invalid MNR value: %s", self.gsi.MNR)
        self.start_offset = DEFAULT_TELETEXT_ROWS
    else:
      self.max_row_count = max_row_count

    # keep track of the end time of the last tti block
    self.last_end_time: typing.Optional[Fraction] = None

    # p_element for use across cumulative subtitles 
    self.cur_p_element = None

  def get_language(self) -> str:
    """Returns the language of the datafile as an RFC 5646 Language Tag
    """
    return self.language

  def get_tti_count(self) -> int:
    """Returns the number of text blocks in the datafile
    """
    return self.tti_count

  def get_fps(self) -> Fraction:
    """Returns the frame rate of the datafile
    """
    return self.fps

  def get_cct(self) -> int:
    """Returns the codepage of the text fields contained in the datafile
    """
    return self.cct

  def get_document(self):
    """Returns the document instance generated from the datafile
    """
    return self.doc

  def is_teletext(self):
    """Returns whether the datafile contains teletext subtitles or open/undefined subtitles
    """
    return ord(self.gsi.DSC) in (0x31, 0x32)

  def get_max_row_count(self):
    """Returns the maximum number of rows
    """
    return self.max_row_count

  def process_tti_block(self, tti_block: bytes):
    """Processes a single TTI block
    """
    if tti_block is None:
      raise ValueError("tti_block should not be None")

    tti = _TTIBlock._make(
      struct.unpack(
        '<BHBBBBBBBBBBBBB112s', tti_block
      )
    )

    LOGGER.debug("Subtitle SN: %s", tti.SN)
    LOGGER.debug("  EBN: %s", tti.EBN)
    LOGGER.debug("  CS: %s", tti.CS)
    LOGGER.debug("  SGN: %s", tti.SGN)
    LOGGER.debug("  JC: %s", tti.JC)
    LOGGER.debug("  VP: %s", tti.VP)

    if 0xEF < tti.EBN < 0xFF:
      # skip user data and reserved blocks
      return

    if not self.is_in_extension:
      self.tti_tf = b''

    self.tti_tf += tti.TF.strip(b'\x8f')

    is_double_height_characters = tf.has_double_height_char(self.tti_tf)

    # continue accumulating if we have an extension block

    if tti.EBN != 0xFF:
      self.is_in_extension = True
      return

    self.is_in_extension = False

    # read TCI and TCO
    try:
      tci = SmpteTimeCode(tti.TCIh, tti.TCIm, tti.TCIs, tti.TCIf, self.get_fps())
      tco = SmpteTimeCode(tti.TCOh, tti.TCOm, tti.TCOs, tti.TCOf, self.get_fps())
    except ValueError:
      LOGGER.error("Invalid TTI timecode")
      return

    LOGGER.debug("  Time in: %s", tci)
    LOGGER.debug("  Time out: %s", tco)

    # compute begin and end time, including program start offset
    begin_time = tci.to_temporal_offset() - self.start_offset
    if begin_time < 0:
      LOGGER.debug("Skipping subtitle because TCI is less than start time")
      return

    end_time = tco.to_temporal_offset() - self.start_offset

    # adjust to avoid overlap unless we are in cumulative mode
    if self.last_end_time is not None and tti.CS == 0x00:
      if begin_time < self.last_end_time:
        LOGGER.warning("Subtitle TCI is less than previous TCO and cumulative mode is inactive; adjusting the former to be equal to the latter")
        begin_time = self.last_end_time

    if end_time <= begin_time:
      LOGGER.error("Subtitle TCO is less than or equal to TCI; omitting subtitle")
      return

    self.last_end_time = end_time


    # create a new subtitle if SN changes and we are not in cumulative mode

    if tti.SN is not self.last_sn and tti.CS in (0x00, 0x01):

      self.last_sn =  tti.SN

      # find the div to which the subtitle belongs, based on SGN

      div_element = self.sgn_to_div_map.get(tti.SGN)

      # create the div if it does not exist

      if div_element is None:
        div_element = model.Div(self.doc)
        self.body.push_child(div_element)
        self.sgn_to_div_map[tti.SGN] = div_element

      # create the p that will hold the subtitle

      self.cur_p_element = model.P(self.doc)

      if tti.JC == 0x01:
        self.cur_p_element.set_style(styles.StyleProperties.TextAlign, styles.TextAlignType.start)
      elif tti.JC == 0x03:
        self.cur_p_element.set_style(styles.StyleProperties.TextAlign, styles.TextAlignType.end)
      else:
        self.cur_p_element.set_style(styles.StyleProperties.TextAlign, styles.TextAlignType.center)

      self.cur_p_element.set_style(
        styles.StyleProperties.LineHeight,
        styles.LengthType(DEFAULT_LINE_HEIGHT_PCT,
        styles.LengthType.Units.pct)
      )

      if self.is_teletext() and not is_double_height_characters:
        font_size = DEFAULT_SINGLE_HEIGHT_FONT_SIZE_PCT
      else:
        font_size = DEFAULT_DOUBLE_HEIGHT_FONT_SIZE_PCT

      self.cur_p_element.set_style(
        styles.StyleProperties.FontSize,
        styles.LengthType(
          font_size,
          styles.LengthType.Units.pct
        )
      )

      safe_area_height =  round(100 - DEFAULT_VERTICAL_SAFE_MARGIN_PCT * 2)
      safe_area_width =  round(100 - DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT * 2)

      # assume that VP < max number of rows/2 means bottom-aligned and otherwise top-aligned
      # probably should offer an option to override this

      if tti.VP < self.get_max_row_count() // 2:
        # top-aligned large region
        
        r_y = DEFAULT_VERTICAL_SAFE_MARGIN_PCT + ((tti.VP - 1) / self.get_max_row_count()) * safe_area_height
        r_height = 100 - DEFAULT_VERTICAL_SAFE_MARGIN_PCT - r_y
        
        region = _get_region_from_model(
          self.doc,
          round(DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT),
          r_y,
          safe_area_width,
          r_height,
          styles.DisplayAlignType.before
        )

      else:

        line_count = tf.line_count(self.tti_tf, is_double_height_characters)
        vp = tti.VP
        line_height = 2 if is_double_height_characters else 1

        r_y = DEFAULT_VERTICAL_SAFE_MARGIN_PCT
        r_height = ((vp + line_count * line_height - 1)/ self.get_max_row_count()) * safe_area_height
        
        region = _get_region_from_model(
          self.doc,
          round(DEFAULT_HORIZONTAL_SAFE_MARGIN_PCT),
          r_y,
          safe_area_width,
          r_height,
          styles.DisplayAlignType.after
        )

      self.cur_p_element.set_region(region)

      div_element.push_child(self.cur_p_element)

    if tti.CS in (0x01, 0x02, 0x03):
      # create a nested span if we are in cumulative mode
      sub_element = model.Span(self.doc)
      self.cur_p_element.push_child(sub_element)
    else :
      sub_element = self.cur_p_element

    sub_element.set_begin(begin_time)
    sub_element.set_end(end_time)

    LOGGER.debug("  TF: %s", self.tti_tf)

    tf.to_model(sub_element, self.is_teletext(), self.get_cct(), self.tti_tf)

    if tti.CS in (0x01, 0x02):
      sub_element.push_child(model.Br(self.doc))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for white space handling during ISD generation'''

# pylint: disable=R0201,C0115,C0116

import unittest
import xml.etree.ElementTree as et
import ttconv.imsc.reader as imsc_reader
from ttconv.isd import ISD
import ttconv.model as model

class LSWPTests(unittest.TestCase):

  def test_lwsp_default(self):
    tree = et.parse('src/test/resources/ttml/lwsp_default.ttml')

    doc = imsc_reader.to_model(tree)

    isd = ISD.from_model(doc, 0)

    p0 = list(isd.iter_regions())[0][0][0][0]

    spans = list(p0)

    self.assertEqual(len(spans), 3)

    self.assertEqual(spans[0][0].get_text(), "hello ")

    self.assertEqual(spans[1][0].get_text(), "my name")

    self.assertEqual(spans[2][0].get_text(), " is Mathilda")

    p1 = list(isd.iter_regions())[0][0][0][1]

    spans = list(p1)

    self.assertEqual(len(spans), 4)

    self.assertEqual(spans[0][0].get_text(), "bonjour")

    self.assertEqual(spans[1][0].get_text(), " mon nom")

    self.assertIsInstance(spans[2], model.Br)

    self.assertEqual(spans[3][0].get_text(), "est")

  def test_lwsp_preserve(self):
    tree = et.parse('src/test/resources/ttml/lwsp_preserve.ttml')

    doc = imsc_reader.to_model(tree)

    isd = ISD.from_model(doc, 0)

    p0 = list(isd.iter_regions())[0][0][0][0]

    spans = list(p0)

    self.assertEqual(len(spans), 3)

    self.assertEqual(spans[0][0].get_text(), "hello ")

    self.assertEqual(spans[1][0].get_text(), " my \nname ")

    self.assertEqual(spans[2][0].get_text(), "is Mathilda")

if __name__ == '__main__':
  unittest.main()
